"""Basketix seasons module"""

from typing import Dict, List, Optional, Tuple, Union

from ..entities import (
    BasketixDate,
    BasketixError,
    LeagueStatus,
    NbaPlayersMotions,
    RankingType,
    Transactions,
)
from ..helpers import DateHelper, uid
from ..tables import BasketixSeasonsTable
from .basketix_weeks import BasketixWeeksHandler


class BasketixSeasonsHandler:
    """Handle basketix seasons"""

    BEGIN_DAY_OF_WEEK: str = "monday"
    DEFAULT_SALARY_CAP = 120
    DEFAULT_DRAFT_PICK_SECOND_DELAY = 5 * 60
    MIN_DRAFT_DELAY_SECONDS = 30
    DEFAULT_RANKING_TYPE = RankingType.POINTS
    DEFAULT_N_SUBSTITUTE = 2
    DEFAULT_POINTS_TABLE = {
        "1": 20,
        "2": 16,
        "3": 13,
        "4": 11,
        "5": 10,
        "6": 9,
        "7": 8,
        "8": 7,
        "9": 6,
        "10": 5,
        "11": 4,
        "12": 3,
        "13": 2,
        "14": 1,
    }
    DEFAULT_BEST_PLAYER_BONUS_POINT = 1
    DEFAULT_BEST_GAME_BONUS_POINT = 1

    def __init__(self, environment: str) -> None:
        """Init basketix season handler"""

        self.basketix_seasons_table = BasketixSeasonsTable(environment)
        self.basketix_weeks_handler = BasketixWeeksHandler(environment)

    def get(self, season_id: str) -> dict:
        """Returns the season corresponding"""

        return self.basketix_seasons_table.query(season_id)[0]

    def get_status(self, season_id: Optional[str] = None, season: Optional[dict] = None) -> str:
        """Returns the status of the season"""

        if not season:
            season_id = season_id if season_id else ""
            season = self.get(season_id)

        if not season["seasonStatus"]["draft"]["isFinished"]:
            return LeagueStatus.DRAFT
        if self._is_regular_season_finished(season.get("regularSeason", {})):
            return LeagueStatus.FINISHED

        return LeagueStatus.REGULAR_SEASON

    def get_number_of_members(self, season_id: str) -> int:
        """Returns the number of users in the season"""

        return len(self.get(season_id)["basketixs"])

    def get_next_season_settings(self, season_id: str) -> dict:
        """Returns the default settings for the next season"""

        season = self.get(season_id)

        return {
            "salaryCap": season["salaryCap"],
            "draftType": season["draft"]["nextDraftType"],
            "isDraftReverse": season["draft"]["nextDraftIsReverse"],
            "nDraftRound": season["draft"]["nRound"],
            "draftPickSecondDelay": season["draft"]["draftPickSecondDelay"],
            "nWeek": season["regularSeason"]["nWeeks"],
            "nSubstitute": season["regularSeason"]["nSubstitute"],
            "rankingType": season["regularSeason"]["rankingType"],
            "pointsTable": season["regularSeason"]["pointsTable"],
        }

    def create_season(
        self,
        league_id: str,
        n_week: int,
        ranking_type: Optional[RankingType],
        draft_order: list,
        n_round_draft: int,
        draft_type: str,
        is_draft_reverse: bool,
        draft_pick_second_delay: Optional[int],
        players: dict,
        salary_cap: Optional[int],
        points_table: Optional[Dict[str, int]],
        next_season_draft_type: str,
        next_season_is_draft_reverse: bool,
        n_substitute: Optional[int],
        best_player_bonus_points: Optional[int],
        best_game_bonus_points: Optional[int],
    ) -> dict:
        """Create a new season"""

        ranking_type = ranking_type if ranking_type else self.DEFAULT_RANKING_TYPE
        salary_cap = salary_cap if salary_cap is not None else self.DEFAULT_SALARY_CAP
        draft_pick_second_delay = draft_pick_second_delay if draft_pick_second_delay else self.MIN_DRAFT_DELAY_SECONDS
        n_substitute = n_substitute if n_substitute is not None else self.DEFAULT_N_SUBSTITUTE
        best_player_bonus_points = (
            best_player_bonus_points if best_player_bonus_points is not None else self.DEFAULT_BEST_PLAYER_BONUS_POINT
        )
        best_game_bonus_points = (
            best_game_bonus_points if best_game_bonus_points is not None else self.DEFAULT_BEST_GAME_BONUS_POINT
        )

        if draft_pick_second_delay < self.MIN_DRAFT_DELAY_SECONDS:
            raise BasketixError("InvalidPickDelay", tokens={"min_delay": self.MIN_DRAFT_DELAY_SECONDS})

        if points_table:
            try:
                [(int(rank), int(pts)) for rank, pts in points_table.items()]
            except ValueError as err:
                raise BasketixError("InvalidPointsTable") from err
        else:
            points_table = self.DEFAULT_POINTS_TABLE

        season = {
            "seasonId": self._generate_season_id(),
            "leagueId": league_id,
            "creationDateTime": DateHelper.now(),
            "basketixIds": draft_order,
            "basketixs": {},
            "salaryCap": salary_cap,
            "draft": {
                "order": self._init_draft(draft_order, is_draft_reverse, n_round_draft),
                "players": players,
                "draftPickSecondDelay": draft_pick_second_delay,
                "draftType": draft_type,
                "isReverse": is_draft_reverse,
                "nRound": n_round_draft,
                "nextDraftType": next_season_draft_type,
                "nextDraftIsReverse": next_season_is_draft_reverse,
            },
            "regularSeason": {
                "rankingType": ranking_type,
                "nWeeks": n_week,
                "nSubstitute": n_substitute,
                "pointsTable": points_table,
                "bonusPoints": {
                    "bestPlayer": best_player_bonus_points,
                    "bestGame": best_game_bonus_points,
                },
            },
            "players": {},
            "seasonStatus": {},
        }

        for basketix_id in draft_order:
            season["basketixs"][basketix_id] = {
                "watchedPalyer": [],
                "transactions": [
                    {
                        "type": Transactions.GAME_CONFIG,
                        "salaryCap": salary_cap,
                        "isSecret": False,
                    }
                ],
            }

        season["seasonStatus"]["draft"] = self._update_draft_status(
            season["draft"]["order"], 0, draft_pick_second_delay
        )

        self.basketix_seasons_table.insert([season])

        return season

    def pick_player(self, season_id: str, pick_number: int, basketix_id: str, player_id: Union[str, None]) -> dict:
        """Pick the player or skip the pick if no player"""

        season = self.basketix_seasons_table.query(season_id)[0]

        draft_status = season["seasonStatus"]["draft"]
        if draft_status["basketixId"] != basketix_id or draft_status["pickNumber"] != pick_number:
            raise BasketixError("CanNotPick")

        draft_pick_second_delay = season["draft"]["draftPickSecondDelay"]

        cost = season["draft"]["players"][player_id]["cost"] if player_id else None
        positions = season["draft"]["players"][player_id]["positions"] if player_id else None

        season = self._pick_player_recursive(
            season, pick_number, basketix_id, player_id, cost, positions, draft_pick_second_delay
        )
        self.basketix_seasons_table.update([season])

        return season["seasonStatus"]["draft"]

    def set_pre_draft_list(
        self, season_id: str, basketix_id: str, pick_number: int, player_ids: list, auto: bool
    ) -> dict:
        """Set the pre-draft list of an user"""

        season = self.basketix_seasons_table.query(season_id)[0]

        if season["draft"]["order"][pick_number - 1]["basketixId"] != basketix_id:
            raise BasketixError("InvalidPickNumber", tokens={"pick_number": pick_number})

        drafted_players = [pick["playerId"] for pick in season["draft"]["order"] if "playerId" in pick]
        already_drafted_players = list(set(drafted_players) & set(player_ids))
        if already_drafted_players:
            raise BasketixError("AlreadyDrafted")

        season["draft"]["order"][pick_number - 1]["predraft"]["choices"] = player_ids
        season["draft"]["order"][pick_number - 1]["predraft"]["auto"] = auto
        update_item = {"seasonId": season_id}
        update_item["draft"] = season["draft"]
        self.basketix_seasons_table.update([update_item])

        return season["draft"]["order"][pick_number - 1]

    def get_roster(self, season_id: str, basketix_id: str, week_id: str) -> dict:
        """Get players positions and costs from basketix rosters"""

        roster = {}
        for player_id, player_history in self._get_players(season_id=season_id).items():
            player_motion = self.get_owner_and_info(week_id, player_history)
            if player_motion and player_motion["basketixId"] == basketix_id:
                roster[player_id] = {
                    "basketixId": basketix_id,
                    "playerId": player_id,
                    "cost": player_motion["cost"],
                    "positions": player_motion["positions"],
                }

        return roster

    def players_in_roster(
        self,
        basketix_id: str,
        week_id: str,
        player_ids: List[str],
        season_id: Optional[str] = None,
        players: Optional[dict] = None,
    ):
        """Returns only the players_ids who are parts of the basketix roster for the weekId"""

        player_ids_in_roster = []
        if not players:
            players = self._get_players(season_id=season_id)

        for player_id in player_ids:
            owner = self.get_owner_and_info(week_id, players[player_id])
            if owner and owner["basketixId"] == basketix_id:
                player_ids_in_roster.append(player_id)

        return player_ids_in_roster

    def get_all_players(self, season_id: str, week_id: str) -> dict:
        """Returns all players with owner"""

        season = self.basketix_seasons_table.query(season_id)[0]
        players = self._get_players(season=season)

        players_with_owner = {}
        for player_id, player_history in players.items():
            player_motion = self.get_owner_and_info(week_id, player_history)
            if player_motion and "basketixId" in player_motion:
                players_with_owner[player_id] = {key: player_motion[key] for key in ["positions", "cost", "basketixId"]}

        return players_with_owner

    def is_season_finished(self, season_id: str) -> bool:
        """Check if season is finished"""

        season = self.basketix_seasons_table.query(season_id)[0]

        is_draft_finished = self._is_draft_finished(season["draft"]["order"])
        is_regular_season_finished = self._is_regular_season_finished(season["regularSeason"])

        return is_draft_finished and is_regular_season_finished

    def get_cap_space(self, transcations: list, with_secrets: bool) -> Tuple[int, Optional[int]]:
        """
        Returns a tuple of :
        - the salary cap based on transactions (with or without secret transactions)
        - the public salary cap if with_secrets = True else None
        """

        salary_cap_transactions = []
        secret_transactions = [0]
        for transaction in transcations:
            if "salaryCap" in transaction and transaction.get("isSecret", False) in (False, with_secrets):
                salary_cap_transactions.append(transaction["salaryCap"])

            if "salaryCap" in transaction and transaction.get("isSecret"):
                secret_transactions.append(transaction["salaryCap"])

        return (
            sum(salary_cap_transactions),
            sum(salary_cap_transactions) - sum(secret_transactions) if with_secrets else None,
        )

    # Deprecated
    def filter_teams_players(self, basketix_teams: List[dict]) -> List[dict]:
        """Delete NBA players who are not in the basketix roster"""

        season_ids = list(set([team["seasonId"] for team in basketix_teams]))

        filtered_teams = []
        for season_id in season_ids:
            teams = [team for team in basketix_teams if team["seasonId"] == season_id]
            season_players = self._get_players(season_id)
            for team in teams:
                filter_team = self._delete_players_not_in_roster(team, season_players)
                filtered_teams.append(filter_team)

        return filtered_teams

    def get_all_next_week_ids(self, week_id: str, season: dict) -> List[str]:
        """Return all the next weeks from <week_id>"""
        try:
            week_ids = season["regularSeason"]["weekIds"]
            return week_ids[
                week_ids.index(
                    week_id,
                ) :
            ]
        except (KeyError, ValueError):
            return []

    def get_n_max_substitute(self, season_id: str) -> int:
        return self.get(season_id=season_id)["regularSeason"]["nSubstitute"]

    # Deprecated
    def _delete_players_not_in_roster(self, team: dict, season_players: dict):
        """Init the position for NBA players who are not in the basketix roster"""

        player_ids = [player_id for player_id, _ in team["players"].items()]
        player_ids_in_roster = self.players_in_roster(
            team["basketixId"], team["weekId"], player_ids, players=season_players
        )
        player_id_not_in_roster = list(set(player_ids) - set(player_ids_in_roster))
        for player_id in player_id_not_in_roster:
            team["players"].pop(player_id)

        return team

    def _init_draft(self, draft_order: list, is_reverse: bool, n_round_draft: int) -> list:
        """Init draft"""

        n_basketix = len(draft_order)
        return [
            self._init_pick(basketix_id, i + 1, n_basketix)
            for i, basketix_id in enumerate(self._get_all_draft_order(draft_order, is_reverse, n_round_draft))
        ]

    def _pick_player_recursive(
        self,
        season: dict,
        pick_number: int,
        basketix_id: str,
        player_id: Optional[str],
        cost: Optional[int],
        positions: Optional[list],
        draft_pick_second_delay: int,
    ) -> dict:
        """Pick a player and the next ones if possible. Update season status"""

        utc_now = DateHelper.now()
        pick = season["draft"]["order"][pick_number - 1]
        pick_update_base = {"engagementDateTime": utc_now, "basketixId": basketix_id, "isPick": True}

        if not player_id or not self._can_pick(basketix_id, season):
            pick.pop("predraft")
            pick.update(pick_update_base)
            return self._next_pick(season, pick_number, draft_pick_second_delay)

        if not self._can_pick_this_player(basketix_id, season, player_id):
            raise BasketixError("CanNotPick")

        pick.pop("predraft")
        pick.update(pick_update_base)
        pick.update({"playerId": player_id, "cost": cost, "positions": positions})

        player_histories = [
            {
                "date": utc_now,
                "type": NbaPlayersMotions.DRAFT,
                "basketixId": basketix_id,
                "cost": cost,
                "positions": positions,
            }
        ]
        season["players"][player_id] = player_histories

        season["basketixs"][basketix_id]["transactions"].append(
            {
                "type": Transactions.DRAFT_PICK,
                "playerId": player_id,
                "salaryCap": -cost,
                "isSecret": False,
            }
        )

        return self._next_pick(season, pick_number, draft_pick_second_delay)

    def _next_pick(self, season: dict, pick_number: int, draft_pick_second_delay: int):
        season["seasonStatus"]["draft"] = self._update_draft_status(
            season["draft"]["order"], pick_number, draft_pick_second_delay
        )

        if season["seasonStatus"]["draft"]["isFinished"]:
            season = self._close_draft(season)
            season = self._init_regular_season(season)
        else:
            next_basketix_id = season["seasonStatus"]["draft"]["basketixId"]
            next_pick_number = season["seasonStatus"]["draft"]["pickNumber"]

            if not self._can_pick(next_basketix_id, season):
                season = self._pick_player_recursive(
                    season, next_pick_number, next_basketix_id, None, None, None, draft_pick_second_delay
                )
                return season

            predraft = season["draft"]["order"][next_pick_number - 1]["predraft"]
            if predraft["auto"]:
                for pick_player_id in predraft["choices"]:
                    if self._can_pick_this_player(next_basketix_id, season, pick_player_id):  # TODO: A tester
                        next_positions = season["draft"]["players"][pick_player_id]["positions"]
                        next_cost = season["draft"]["players"][pick_player_id]["cost"]
                        season = self._pick_player_recursive(
                            season,
                            next_pick_number,
                            next_basketix_id,
                            pick_player_id,
                            next_cost,
                            next_positions,
                            draft_pick_second_delay,
                        )
                        break

        return season

    def _update_draft_status(self, draft: list, actual_pick_number: int, draft_pick_second_delay: int) -> dict:
        """Returns if it's the turn of a specific player to pick a player during the draft"""

        is_finished = self._is_draft_finished(draft)
        draft_status = {"isFinished": is_finished}
        if not is_finished:
            # Next pick number
            draft_status = {
                **draft_status,
                "basketixId": draft[actual_pick_number]["basketixId"],
                "pickNumber": draft[actual_pick_number]["pickNumber"],
                "round": draft[actual_pick_number]["round"],
                "roundPick": draft[actual_pick_number]["roundPick"],
                "limitDateTime": DateHelper.now_delta(seconds=draft_pick_second_delay),
            }

        return draft_status

    def _init_regular_season(self, season: dict) -> dict:
        """Init a regular season after the draft is finished"""

        if self._is_draft_finished(season["draft"]["order"]):
            season_week_ids = self._get_season_week_ids(season["regularSeason"]["nWeeks"])
            season["regularSeason"] = {
                **season["regularSeason"],
                "beginWeekId": season_week_ids[0],
                "endWeekId": season_week_ids[-1],
                "weekIds": season_week_ids,
            }
            season["seasonStatus"]["regularSeason"] = self._update_regular_season_status(season["regularSeason"])

        return season

    def _update_regular_season_status(self, regular_season: dict) -> dict:
        """Update regular season status"""

        return {"isFinished": self._is_regular_season_finished(regular_season)}

    def _is_draft_finished(self, draft: list) -> bool:
        """Check if draft is finished"""
        return draft[-1]["isPick"]

    def _is_regular_season_finished(self, regular_season: dict) -> bool:
        """Check if regular season is finished"""

        if "weekIds" in regular_season:
            return self.basketix_weeks_handler.are_all_finished(regular_season["weekIds"])

        return False

    def get_owner_and_info(self, week_id: str, history: list) -> Optional[dict]:
        """
        Finds basketix owner for a player motion history and a <week_id>
        and returns players main info (cost, positions, owner)
        """

        if "lastWeekId" in history[-1] and "lastWeekId" in history[-1] and week_id > history[-1]["lastWeekId"]:
            return None

        for motion in reversed(history):
            if (
                "lastWeekId" in motion
                and "firstWeekId" in motion
                and motion["firstWeekId"] <= week_id <= motion["lastWeekId"]
            ):
                return motion

            if "lastWeekId" in motion and "firstWeekId" not in motion and week_id <= motion["lastWeekId"]:
                return motion

            if not "lastWeekId" in motion and "firstWeekId" in motion and motion["firstWeekId"] <= week_id:
                return motion

        if history[0]["type"] == "Draft" and (not "lastWeekId" in history[0] or week_id <= history[0]["lastWeekId"]):
            return history[0]

        return None

    def _get_players(self, season_id: Optional[str] = None, season: Optional[dict] = None) -> dict:
        """Get players positions and costs from all basketix rosters"""

        if season_id:
            season = self.basketix_seasons_table.query(season_id)[0]
        elif not season:
            raise BasketixError("InvalidSeason")

        return season["players"]

    def _close_draft(self, season: dict):
        """Close draft when it is finish. Delete none drafted player"""

        season["draft"].pop("players", season)

        return season

    def _generate_season_id(self) -> str:
        """Generate a season id"""

        return uid()

    def _get_all_draft_order(self, draft_order: list, is_reverse: bool, n_round_draft: int) -> list:
        """Returns ordered basketix id list of the draft"""

        draft_orders = [draft_order for _ in range(n_round_draft)]

        if is_reverse:
            for _round, _ in enumerate(draft_orders):
                if _round % 2 == 1:
                    draft_orders[_round] = list(reversed(draft_orders[_round]))

        return [_round for draft_round in draft_orders for _round in draft_round]

    def _init_pick(self, basketix_id: str, position: int, n_basketix: int):
        """Init a draft pick"""

        return {
            "basketixId": basketix_id,
            "isPick": False,
            "pickNumber": position,
            "round": ((position - 1) // n_basketix) + 1,
            "roundPick": position - (n_basketix * (((position - 1) // n_basketix))),
            "predraft": {"auto": True, "choices": []},
        }

    def _can_pick(self, basketix_id: str, season: dict) -> bool:
        """Check if basketix has cap space i.e he can draft a player"""

        return self.get_cap_space(season["basketixs"][basketix_id]["transactions"], with_secrets=True)[0] > 0

    # TODO: A tester
    def _can_pick_this_player(self, basketix_id: str, season: dict, player_id: str) -> bool:
        """
        Check if basketix <basketix_id> can pick player <player_id> dependending
        of the draft history and basketix cap space
        """
        already_drafted_players = [player_id for player_id, _ in season["players"].items()]
        if player_id in already_drafted_players:
            return False

        cap_space, _ = self.get_cap_space(season["basketixs"][basketix_id]["transactions"], with_secrets=True)
        if cap_space == 0 or cap_space < season["draft"]["players"][player_id]["cost"]:
            return False

        return True

    def _get_season_week_ids(self, n_week: int) -> list:
        """Returns the list of weeks id for the regular season based on now"""

        first_week = BasketixDate.tomorrow().week.next()
        return [first_week.next(i).id for i in range(n_week)]



"""NBA players table"""

from .dynamodb_table import DynamoDBTable
from ..models import NbaPlayerModel

class NbaPlayersTable(DynamoDBTable[NbaPlayerModel]):
    def __init__(self, environment: str):
        DynamoDBTable.__init__(self, 'nba-players', environment, 'playerId')



"""NBA games module"""

from typing import List, Optional, Dict

from boto3.dynamodb.conditions import Key

from ..tables import NbaGamesTable
from ..entities import BasketixWeek, BasketixDate
from ..models import NbaGameModel, NbaGameStatus

class NbaGamesHandler:
    """Handle NBA games."""

    def __init__(self, environment: str):
        """Init NBA games handler."""

        self.nba_games_table = NbaGamesTable(environment=environment)

    def get_game_info(self, game_id: str) -> NbaGameModel:
        """Get game information from database."""

        game = self.nba_games_table.query(game_id)[0]

        return game

    def get_game_date(self, game_id: str) -> str:
        """Returns date of a game"""

        return self.get_game_info(game_id=game_id)['gameDate']

    def get_all_games_of_day(self, date: str) -> List[NbaGameModel]:
        """Get all games for a specific date."""

        week = BasketixWeek.from_date(date)

        return self.nba_games_table.query(Key('weekId').eq(week.id) & Key('gameDate').eq(date), index_name='nba-games-table-secondary-index')

    def get_all_finished_games_of_day(self, date: str) -> List[NbaGameModel]:
        """Get all finished games for a specific date."""

        return [game for game in self.get_all_games_of_day(date) if self.is_finished(game=game)]

    def get_all_game_ids_of_day(self, date: str) -> List[str]:
        """Get all game ids for a specific date."""

        return [game['gameId'] for game in self.get_all_games_of_day(date=date)]

    def is_finished(self, game_id: Optional[str] = None, game: Optional[NbaGameModel] = None) -> bool:
        """Returns if game is finish or not"""

        if not game:

            if not game_id:
                raise Exception('One of game_id or game should be defined')

            game = self.get_game_info(game_id)

        return game['gameStatus'] == NbaGameStatus.FINISHED

    def get_all_games_of_week_for_team(self, team_id: str, week_id: str) -> List[NbaGameModel]:
        """Returns all games for a specific week and team"""

        games = self.get_all_games_of_week(week_id)

        return [game for game in games if team_id in game['teamIds']]

    def get_game_info_for_team(self, game: NbaGameModel, team_id: str) -> dict:
        """Returns game basic info with the scope of team_id"""

        return {
            'gameId': game['gameId'],
            'gameDate': game['gameDate'],
            'gameStatus': game['gameStatus'],
            'startTimeUTC': game['gameId'],
            'isHome': [team['location'] for team in game['teams'] if team['teamId'] == team_id][0] == 'home',
            'isAway': [team['location'] for team in game['teams'] if team['teamId'] == team_id][0] == 'away',
            'opponentId': [_id for _id in game['teamIds'] if _id != team_id][0],
        }

    def get_all_games_of_week(self, week_id: str) -> List[NbaGameModel]:
        """Get all game for a specific week."""

        return self.nba_games_table.query(Key('weekId').eq(week_id), index_name='nba-games-table-secondary-index')

    def get_last_finished_game(self, team_ids: List[str], max_occurrence: int = 7) -> Dict[str, NbaGameModel]:
        """Returns for each team id the last finished game, if any"""

        last_game_by_team = {}
        team_ids_set = set(team_ids)

        occurrence = 0
        cursor_date = BasketixDate.today()
        while occurrence <= max_occurrence and team_ids_set - set(last_game_by_team):
            games = self.get_all_finished_games_of_day(date=cursor_date.date)
            last_game_by_team = {
                **last_game_by_team,
                **{team_id: game for game in games for team_id in game['teamIds'] if team_id not in last_game_by_team}
            }

            cursor_date = cursor_date.day_before
            occurrence += 1

        return last_game_by_team



"""Basketix teams table"""

from .dynamodb_table import DynamoDBTable
from ..models import BasketixTeamsModel

class BasketixTeamsTable(DynamoDBTable[BasketixTeamsModel]):
    def __init__(self, environment: str):
        DynamoDBTable.__init__(self, 'basketix-teams', environment,  'weekId', 'teamId')



"""WeekId helper module"""

from typing import List, Optional

from ..entities import BasketixError
from ..helpers import DateHelper
from ..tables import BasketixWeeksTable


class BasketixWeeksHandler:
    def __init__(self, environment: str) -> None:
        self.basketix_weeks_table = BasketixWeeksTable(environment)

    def are_all_finished(self, week_ids: List[str]) -> bool:
        """Returns if all week_id in week_ids are finished"""

        for week_id in week_ids:
            week_res = self.basketix_weeks_table.query(week_id)
            if not week_res:
                return False
            all_days_finished = week_res[0].get("isFinished", False)
            if not all_days_finished:
                return False

        return True

    def get_start_time_utc(self, week_id: str) -> Optional[str]:
        """Returns the start time (UTC) for a week id"""

        weeks = self.basketix_weeks_table.query(week_id)
        if weeks:
            return weeks[0].get("startTimeUTC")

        return None

    def is_week_started(self, week_id: Optional[str] = None, week: Optional[dict] = None) -> bool:
        """Returns if the week is started or not"""

        week = self._get_week(week_id, week)
        if not week:
            return False

        start_time_utc: Optional[str] = week.get("startTimeUTC")
        if not start_time_utc:
            return False

        utc_now = DateHelper.now()

        return utc_now >= start_time_utc

    def is_week_finished(self, week_id: Optional[str] = None, week: Optional[dict] = None) -> bool:
        """Returns if the week is finished or not"""

        week = self._get_week(week_id, week)
        if not week:
            return False

        is_finished = week.get("isFinished")

        if is_finished is None:
            return False

        return is_finished

    def is_week_live(self, week_id: Optional[str] = None, week: Optional[dict] = None) -> bool:
        """Returns if the week is live or not"""

        week = self._get_week(week_id, week)
        if not week:
            return False

        return self.is_week_started(week=week) and not self.is_week_finished(week=week)

    def get_week_status(self, week_id: str) -> dict:
        """Returns if week is starter, finished and live"""

        weeks = self.basketix_weeks_table.query(week_id)
        if not weeks:
            return {
                "weekId": week_id,
                "isStarted": False,
                "isFinished": False,
                "isLive": False,
            }

        week = weeks[0]
        return {
            "weekId": week_id,
            "isStarted": self.is_week_started(week=week),
            "isFinished": self.is_week_finished(week=week),
            "isLive": self.is_week_live(week=week),
            "startTimeUTC": week["startTimeUTC"],
        }

    def get_weeks_status(self, week_ids: List[str]) -> List[dict]:
        """Returns if weeks are starter, finished and live"""

        return [self.get_week_status(week_id) for week_id in week_ids]

    def _get_week(self, week_id: Optional[str], week: Optional[dict] = None) -> Optional[dict]:
        """Queries week if necessary and returns it"""
        if not week_id and not week:
            raise BasketixError("InvalidParameters")

        if not week:
            weeks = self.basketix_weeks_table.query(week_id)
            if not weeks:
                return None
            week = weeks[0]

        return week



"""Basketix seasons table"""

from .dynamodb_table import DynamoDBTable

class BasketixSeasonsTable(DynamoDBTable):
    def __init__(self, environment):
        DynamoDBTable.__init__(self, 'basketix-seasons', environment, 'seasonId')



"""Basketix free agencies table"""

from .dynamodb_table import DynamoDBTable

class BasketixFreeAgenciesTable(DynamoDBTable):
    def __init__(self, environment):
        DynamoDBTable.__init__(self, 'basketix-free-agencies', environment, 'seasonId', 'freeAgentId')



"""Week id module"""

from datetime import datetime
from typing import List, Type

from isoweek import Week
from typing_extensions import Self

from ..helpers.date_helper import DateHelper
from .basketix_season import BasketixSeason
from .game_id import GameId


class BasketixWeek:
    """Basketix week class"""

    SEPARATOR = "_"
    WEEK_MODE = {
        "REGULAR": "REGULAR",
        "PO": "PLAYOFF",
    }

    def __init__(self, value: str) -> None:
        self._value = value

    def __repr__(self) -> str:
        return f"Basketix week: {self._value}"

    def __str__(self) -> str:
        return self._value

    def __eq__(self, other):
        return self.first_day == other.first_day and self.mode == other.mode

    def __gt__(self, other):
        return self.first_day > other.first_day

    def __ge__(self, other):
        return self.first_day >= other.first_day

    def __lt__(self, other):
        return self.first_day < other.first_day

    def __le__(self, other):
        return self.first_day <= other.first_day

    def __call__(self):
        return self.id

    @property
    def id(self) -> str:
        """Returns week id"""

        return self._value

    @property
    def year(self) -> int:
        """Returns week year"""

        return int(self._value.split(self.SEPARATOR)[0])

    @property
    def week_number(self) -> int:
        """Returns week number"""

        return int(self._value.split(self.SEPARATOR)[1])

    @property
    def mode(self) -> str:
        """Returns week mode"""

        splitted = self._value.split(self.SEPARATOR)
        if len(splitted) == 3:
            return self.WEEK_MODE[splitted[2]]

        return self.WEEK_MODE["REGULAR"]

    @property
    def days(self) -> List[str]:
        """Returns all days of a week."""

        return [day.strftime(DateHelper.DATE_ISO_FORMAT) for day in Week(self.year, self.week_number).days()]

    @property
    def first_day(self) -> str:
        """Returns first days of a week as string"""

        return self.days[0]

    @property
    def last_day(self) -> str:
        """Returns last days of a week as string"""

        return self.days[-1]

    @property
    def season(self) -> BasketixSeason:
        """Returns all days of a week."""

        first_day_of_week = self.days[0]

        return BasketixSeason.from_date(first_day_of_week)

    def next(self, delta=1) -> Self:
        """Returns next delta week"""

        date = DateHelper.delta_days(self.first_day, days=7 * delta)

        return self.from_date(date=date)

    def previous(self, delta=1) -> Self:
        """Returns previous delta week"""

        date = DateHelper.delta_days(self.first_day, days=-7 * delta)

        return self.from_date(date=date)

    def to_playoff(self) -> Self:
        new_id = self.id + self.SEPARATOR + "PO"
        return self.__class__(new_id)

    @classmethod
    def from_date(cls: Type["Self"], date: str) -> Self:
        """Parses date and returns Week instance"""

        parse_date = datetime.strptime(date, DateHelper.DATE_ISO_FORMAT)
        year, week_number, _ = parse_date.isocalendar()

        week_id = f"{year}_{week_number:02d}"

        return cls(value=week_id)

    @classmethod
    def from_date_and_game_id(cls: Type["Self"], date: str, game_id: GameId) -> Self:
        """Parses date and returns Week instance"""

        week = cls.from_date(date=date)

        if game_id.is_playoff:
            week = week.to_playoff()

        return week



"""Basketix error module"""

import os
import traceback
from typing import Optional

from .language import DEFAULT_LANGUAGE

ENVIRONMENT = os.getenv('ENVIRONMENT', 'test')

class BasketixError(Exception):
    """Basketix error class"""

    ERROR_MESSAGES = {
        "UnknownError": {
            "en": "Unknown error",
            "fr": "Erreur inconnue",
        },
        "Unauthorized": {
            "en": "Not authorized",
            "fr": "Non autorisé",
        },
        "InvalidParameters": {
            "en": "Invalid parameters : { parameters_error }",
            "fr": "Paramètres non valides : { parameters_error }",
        },
        "NotFreeAgent": {
            "en": "The player is not a free agent",
            "fr": "Le joueur n'est pas un agent libre",
        },
        "InvalidBid_TooSmall": {
            "en": "Player is more expensive ({ player_cost } M$) than your bid ({ cost } M$)",
            "fr": "Le joueur est plus cher ({ player_cost } M$) que votre offre ({ cost } M$)",
        },
        "InvalidBid_CutRecently": {
            "en": "You cut the player for less than { min_days_after_cut } days",
            "fr": "Vous avez coupé le joueur il y a moins de { min_days_after_cut } jours",
        },
        "InvalidBid_CapSpace": {
            "en": "You do not have enough cap space ({ cap_space } M$) for this bid ({ cost } M$)",
            "fr": "Vous n'avez pas assez de cap space ({ cap_space } M$) pour cette offre ({ cost } M$)",
        },
        "InvalidBid_LowerThanPrevious": {
            "en": "Your new bid must be higher than the previous one: { old_bid_cost } M$",
            "fr": "Votre nouvelle offre doit être supérieure à l'ancienne : { old_bid_cost }  M$",
        },
        "InvalidWithdraw": {
            "en": "You have not bid for this player",
            "fr": "Vous n'avez pas enchéri pour ce joueur",
        },
        "InvalidOwner": {
            "en": "You do not own the player",
            "fr": "Vous n'êtes pas propriétaire du joueur",
        },
        "InvalidPickDelay": {
            "en": "The draft pick delay should be at least { min_delay } seconds.",
            "fr": "Le délai de sélection doit être d'au moins { min_delay } secondes.",
        },
        "AlreadyInLeague": {
            "en": "You are already in the league",
            "fr": "Vous êtes déjà dans la ligue",
        },
        "AccessToLeagueDenied": {
            "en": "Access to the league denied",
            "fr": "Accès à la ligue refusé",
        },
        "UnknownDraftType": {
            "en": "Unknown draft type",
            "fr": "Type de draft inconnu",
        },
        "InvalidPickNumber": {
            "en": "Pick number { pick_number} is not your pick",
            "fr": "Le choix numéro { pick_number} n'est pas votre choix",
        },
        "AlreadyDrafted": {
            "en": "Some players have already been drafted",
            "fr": "Certains joueurs ont déjà été draftés",
        },
        "CanNotPick": {
            "en": "You can not pick the player",
            "fr": "Vous ne pouvez pas sélectionner le joueur",
        },
        "InvalidSeason": {
            "en": "Can not get season",
            "fr": "Impossible d'obtenir la saison",
        },
        "TeamAlreadyExists": {
            "en": "A team already exists",
            "fr": "A team already exists",
        },
        "InvalidPositionsTeam": {
            "en": "Player positions not allowed to form a team",
            "fr": "Les positions des joueurs ne sont pas autorisées à former une équipe",
        },
        "InvalidEmail": {
            "en": "{ email } already used",
            "fr": "{ email } déjà utilisée",
        },
        "InvalidConfirmationCode": {
            "en": "Confirmation code is not valid",
            "fr": "Confirmation code n'est pas valide",
        },
        "PlayerWithoutCost": {
            "en": "Player { player_id } does not have a cost",
            "fr": "Le joueur { player_id } n'a pas de coût",
        },
        "InvalidEmailFormat": {
            "en": "Invalid email format",
            "fr": "Format d'email invalide",
        },
        "InvalidPasswordFormat": {
            "en": "Invalid password policy",
            "fr": "Politique de mot de passe invalide",
        },
        "InvalidPasswordConfirmation": {
            "en": "Non-identical passwords",
            "fr": "Les mots de passe ne sont pas identiques",
        },
        "NotLeagueMember": {
            "en": "You are not a member of this league",
            "fr": "Vous n'êtes pas membre de cette ligue",
        },
        "NotFinishedSeason": {
            "en": "At least one season not completed",
            "fr": "Au moins une saison non terminée",
        },
        "StartedWeek": {
            "en": "This week began",
            "fr": "Cette semaine a commencé",
        },
        "NotInRoster": {
            "en": "Some players are not in your roster",
            "fr": "Certains joueurs ne font pas partie de votre effectif",
        },
        "FinishedDraft": {
            "en": "Draft is finished: no free agents found",
            "fr": "La draft est terminée : aucun agent libre n'a été trouvé",
        },
        "PlayerMustHaveOnePosition": {
            "en": "A player can have only one position",
            "fr": "Un joueur ne peut avoir qu'une seule position",
        },
        "InvalidPointsTable": {
            "en": "Invalid points table",
            "fr": "Tableau de points invalide",
        },
    }


    def __init__(self, error_code: str, tokens: Optional[dict] = None, inner_exception: Optional[Exception] = None, status_code=400):
        """Init the basketix error."""
        Exception.__init__(self)
        self.error_code = error_code
        self.error_message = None
        self.tokens: dict = tokens if tokens else {}
        self.status_code = status_code
        self._inner_exception = inner_exception

    def __str__(self):
        msg = f"BasketixError : {self.error_code}"
        if self._inner_exception is not None:
            trb = '\n'.join(traceback.format_exception(self._inner_exception.__class__,
                                                       self._inner_exception,
                                                       self._inner_exception.__traceback__))
            msg = msg + '\n' + trb
        return msg

    def get(self, language: Optional[str]) -> dict:
        error = {
            'type': 'BasketixError',
            'code': self.error_code,
            'message': self._error_message(language),
        }
        if ENVIRONMENT != 'prod':
            error['traceback'] = str(self)

        return error

    def _error_message(self, language: Optional[str]) -> str:
        language = language if language else DEFAULT_LANGUAGE
        messages = self.ERROR_MESSAGES[self.error_code]
        message = messages[language] if language in messages else messages[DEFAULT_LANGUAGE]

        for key, value in self.tokens.items():
            message = message.replace(f'{{ {key} }}', str(value))

        return message



"""NBA games table"""

from .dynamodb_table import DynamoDBTable

class NbaTeamsTable(DynamoDBTable):
    def __init__(self, environment):
        DynamoDBTable.__init__(self, 'nba-teams', environment, 'teamId')



"""Basketball court handler"""

from ..models import BasketballCourtSettings, BasketixId, SeasonId
from ..tables import BasketixSeasonsTable


class BasketballCourtHandler:

    BLACK = "#000000"
    DEFAULT_COURT_COLOR = "#e6be91"
    DEFAULT_LINE_WIDTH = 3

    def __init__(self, environment: str) -> None:
        self.basketix_seasons_table = BasketixSeasonsTable(environment)


    def get(self, season_id: SeasonId, basketix_id: BasketixId) -> BasketballCourtSettings:
        season = self.basketix_seasons_table.get(season_id)

        if "courtSettings" in season["basketixs"][basketix_id]:
            return season["basketixs"][basketix_id]["courtSettings"]

        return self.default()

    def upsert(self, season_id: SeasonId, basketix_id: BasketixId, court_settings: BasketballCourtSettings) -> BasketballCourtSettings:
        """Update or insert court setting of a user"""
        season = self.basketix_seasons_table.get(season_id)

        court = self.default()
        if "courtSettings" in season["basketixs"][basketix_id]:
            court: BasketballCourtSettings = season["basketixs"][basketix_id]["courtSettings"]

        updated_court: BasketballCourtSettings = {**court, **court_settings} # type: ignore
        season["basketixs"][basketix_id]["courtSettings"] = updated_court
        self.basketix_seasons_table.update([season])

        return updated_court

    def default(self) -> BasketballCourtSettings:
        return {
            "backgroundColor": self.DEFAULT_COURT_COLOR,
            "courtColor": self.DEFAULT_COURT_COLOR,
            "linesColor": self.BLACK,
            "linesWidth": self.DEFAULT_LINE_WIDTH,
            "paintColor": self.DEFAULT_COURT_COLOR,
            "sidePaintColor": self.DEFAULT_COURT_COLOR,
            "threePointsColor": self.DEFAULT_COURT_COLOR,
        }



"""AWS helpers module"""

import json

def get_sns_message(event: dict):
    """Returns parsed SNS message"""
    try:
        return json.loads(event['Records'][0]['Sns']['Message'])
    except KeyError:
        return event



from .basketix_date import BasketixDate
from .basketix_error import BasketixError
from .basketix_season import BasketixSeason
from .basketix_week import BasketixWeek
from .game_id import GameId
from .injury import InjuryType
from .language import DEFAULT_LANGUAGE, LANGUAGES
from .league_status import LeagueStatus
from .nba_player_motions import NbaPlayersMotions
from .ranking_type import RankingType
from .transactions import Transactions



"""Basketix weeks table"""

from .dynamodb_table import DynamoDBTable

class BasketixWeeksTable(DynamoDBTable):
    def __init__(self, environment):
        DynamoDBTable.__init__(self, 'basketix-weeks', environment, 'weekId')



"""Date helper"""

from datetime import datetime, timedelta
from typing import Type

from dateutil.parser import parse
from dateutil import tz
from pytz import utc

class DateHelper:
    """Date helper class"""

    US_EASTERN_TIMEZONE = tz.gettz('US/Eastern')
    DATE_ISO_FORMAT = '%Y-%m-%d'
    UTC = utc

    @classmethod
    def now(cls: Type['DateHelper']) -> str:
        """Returns UTC ISO-formatted now"""

        return datetime.now().astimezone(utc).isoformat()

    @classmethod
    def delta_days(cls: Type['DateHelper'], date: str, days: int) -> str:
        """Returns date + delta ISO-formatted date"""

        _date = cls.parse_ISO_date(date)
        new_date = _date + timedelta(days=days)

        return cls.to_ISO_date(new_date)

    @classmethod
    def now_delta(cls: Type['DateHelper'], days=0, seconds=0, microseconds=0,
            milliseconds=0, minutes=0, hours=0, weeks=0) -> str:
        """Returns ISO-formatted UTC now + delta"""

        utc_now = cls.parse(cls.now())
        new_date = utc_now + timedelta(
            days=0,
            seconds=0,
            microseconds=0,
            milliseconds=0,
            minutes=0,
            hours=0,
            weeks=0
        )

        return new_date.isoformat()

    @classmethod
    def parse(cls: Type['DateHelper'], datetime_str: str) -> datetime:
        """Parses datetime"""

        return parse(datetime_str)

    @classmethod
    def to_ISO_date(cls: Type['DateHelper'], date: datetime) -> str:
        """Returns ISO-formatted date"""

        return date.strftime(cls.DATE_ISO_FORMAT)

    @classmethod
    def parse_ISO_date(cls: Type['DateHelper'], date: str) -> datetime:
        """Parses ISO-formatted date and returns datetime object"""

        return datetime.strptime(date, cls.DATE_ISO_FORMAT)




"""Basketix user table"""

from .dynamodb_table import DynamoDBTable

class BasketixUsersTable(DynamoDBTable):
    def __init__(self, environment):
        DynamoDBTable.__init__(self, 'basketix-users', environment, 'basketixId', 'email')



"""NBA player injury module"""

from enum import Enum


class InjuryType(str, Enum):
    PROBABLE = "PROBABLE"
    DOUBTFUL = "DOUBTFUL"
    QUESTIONABLE = "QUESTIONABLE"
    OUT_NEXT_GAME = "OUT_NEXT_GAME"
    OUT_LAST_GAME = "OUT_LAST_GAME"
    OUT = "OUT"
    OUT_INDEFINITELY = "OUT_INDEFINITELY"
    OUT_FOR_SEASON = "OUT_FOR_SEASON"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def parse(cls, injury_text) -> "InjuryType":
        """Converts Basket-Reference note into a InjuryType"""
        raw_injury_type, injury_detail = injury_text.split(" - ")

        injury_detail = injury_detail.upper()
        raw_injury_type = raw_injury_type.upper()

        if "OUT FOR SEASON" in raw_injury_type:
            return cls.OUT_FOR_SEASON

        if "PROBABLE" in injury_detail:
            return InjuryType.PROBABLE

        if "QUESTIONABLE" in injury_detail:
            return InjuryType.QUESTIONABLE

        if "DOUBTFUL" in injury_detail:
            return InjuryType.DOUBTFUL

        if "DID NOT PLAY IN" in injury_detail:
            return InjuryType.OUT_LAST_GAME

        if "OUT FOR" in injury_detail:
            return InjuryType.OUT_NEXT_GAME

        if "INDEFINITELY" in injury_detail:
            return InjuryType.OUT_INDEFINITELY

        if "WITHOUT A TIMETABLE TO RETURN" in injury_detail:
            return InjuryType.OUT_INDEFINITELY

        if "NO TIMETABLE TO RETURN" in injury_detail:
            return InjuryType.OUT_INDEFINITELY

        if "OUT" in raw_injury_type:
            return InjuryType.OUT

        return InjuryType.UNKNOWN



from .basketix_season import BasketixSeason


class GameId:

    def __init__(self, value: str) -> None:
        self._value = value

    def __repr__(self) -> str:
        return self._value

    def __eq__(self, other):
        return self._value == other._value

    @property
    def league(self) -> int:
        return int(self._value[0:2])

    @property
    def league_enum(self) -> str:
        if self.league == 0:
            return "NBA"

        raise ValueError(f"Unknown league {self.league}")

    @property
    def stage(self) -> int:
        return int(self._value[2])

    @property
    def stage_enum(self) -> str:
        if self.stage == 2:
            return "REGULAR"
        if self.stage == 3:
            return "ALL_STAR"
        if self.stage == 4:
            return "PLAYOFF"

        raise ValueError(f"Unknown stage {self.stage}")

    @property
    def is_playoff(self) -> int:
        return self.stage == 4

    @property
    def season(self) -> int:
        return int(self._value[3:5])

    @property
    def season_basketix(self) -> BasketixSeason:
        first_year_of_the_season = int(f"20{self.season}") - 1
        return BasketixSeason(f"{first_year_of_the_season}-{self.season}")

    @property
    def index(self) -> int:
        return int(self._value[5:])
